from consts import *
import Utils
import Live
from Clip import *
from Track import *
from Scene import *

class Dispatcher:
    """
    General request dispatcher.
    """
    def __init__(self,controller,logger):
        self.controller = controller
        self.logger = logger
        self.handlers = {}
        self.handlers[CLIP_TRIGGER_CHANNEL] = ClipTriggerHandler(CLIP_TRIGGER_CHANNEL, self.controller, self.logger)
        self.handlers[CLIP_DATA_CHANNEL] = ClipDataHandler(CLIP_DATA_CHANNEL, self.controller,self.logger)
        
        clipMonitorHandler = ClipMonitorHandler(CLIP_MONITOR_CHANNEL, CLIP_STOP_MONITOR_CHANNEL, self.controller,self.logger)
        self.handlers[CLIP_MONITOR_CHANNEL] = clipMonitorHandler
        self.handlers[CLIP_STOP_MONITOR_CHANNEL] = clipMonitorHandler
        
        trackDataHandler = TrackDataHandler(GENERAL_REQUEST_CHANNEL, self.controller, self.logger)
        trackMonitorHandler = TrackMonitorHandler(GENERAL_REQUEST_CHANNEL, self.controller, self.logger)
        
        sceneDataHandler = SceneDataHandler(GENERAL_REQUEST_CHANNEL, self.controller, self.logger)
        sceneMonitorHandler = SceneMonitorHandler(GENERAL_REQUEST_CHANNEL, self.controller, self.logger)
        sceneTriggerHandler = SceneTriggerHandler(GENERAL_REQUEST_CHANNEL, self.controller, self.logger)
        
        self.handlers[GENERAL_REQUEST_CHANNEL] = [trackDataHandler,trackMonitorHandler,sceneDataHandler,sceneMonitorHandler,sceneTriggerHandler]
        
    def build_midi_map(self,script_handle,midi_map_handle):
        for channel, handler in self.handlers.iteritems():
            if isinstance(handler,list):
                for subhandler in handler:
                    subhandler.build_midi_map(script_handle,midi_map_handle)
            else:
                handler.build_midi_map(script_handle,midi_map_handle)
    
    def handle_note(chan,note,velocity):
        self.logger.log("Note " + str(chan) + " : " + str(note) + " : " + note(velocity))
        
    def handle_cc(self,chan,num,val):
        self.logger.log("CC " + str(chan) + " : " + str(num) + " : " + str(val))

        handler = self.handlers[chan]
        if handler:
            if isinstance(handler,list):
                for subhandler in handler:
                    if subhandler.can_handle(chan,num,val):
                        data = subhandler.handle_cc(chan,num,val)
                        break
            else:
                data = handler.handle_cc(chan,num,val)
            if data:
                self.controller.sendSysex(data)
        else:
            strTuple = self.translate_string("Unknown")
            self.controller.sendSysex(strTuple)
