autowatch = 1;
inlets = 2;
setinletassist(1,"sysex data");
setinletassist(0,"commands");
outlets = 5;
setoutletassist(4,"scene change events");
setoutletassist(3,"clip change events");
setoutletassist(2,"track change events");
setoutletassist(1,"song change events");
setoutletassist(0,"requests");

// DEBUGGING FLAG
DEBUG = false;

// Incoming message types. Same as in consts.py in the APIMidi python code.
CLIP_DATA  = 1;
EMPTY_CLIP = 2;
CLIP_IS_TRIGGERED = 3;
CLIP_IS_PLAYING = 4;
CLIP_NAME = 5;
CLIP_IS_LOOPING = 6;

TRACK_DATA = 10;
NO_TRACK_DATA = 11;
TRACK_NAME = 12;
TRACK_IS_ARMED = 13;
TRACK_IS_MUTED = 14;
TRACK_IS_SOLOED = 15;

DEVICE_DATA = 20;
NO_DEVICE_DATA = 21;

SCENE_DATA = 30;
NO_SCENE_DATA = 31;
SCENE_NAME = 32;

CUEPOINT_DATA = 40;
NO_CUEPOINT_DATA = 41;

MIXER_DATA = 50;
NO_MIXER_DATA = 51;

// Outgoing messages
GET_CLIP_CHANNEL = 2;
MONITOR_CLIP_CHANNEL = 3;
STOP_MONITOR_CLIP_CHANNEL = 3;

GET_SONG_CHANNEL = 5;
GET_SONG_PARAM = 0;

GET_TRACK_CHANNEL = 5;
MONITOR_TRACK_CHANNEL = 5;
GET_TRACK_PARAM = 1;
MONITOR_TRACK_PARAM = 2; 
STOP_MONITOR_TRACK_PARAM = 3; 

GET_SCENE_CHANNEL = 5;
MONITOR_SCENE_CHANNEL = 5;
GET_SCENE_PARAM = 10;
MONITOR_SCENE_PARAM = 11; 
STOP_MONITOR_SCENE_PARAM = 12; 

GET_MIXER_CHANNEL = 5;
GET_MIXER_PARAM = 20;

// SONG
function Song()
{
	this.tracks = [];
	this.scenes = [];
}

Song.prototype.setSize = function(tracks,clips)
{
	this.clear();
	
	if(DEBUG)
	{
		post("! Creating " + tracks + " tracks and " + clips + " clips.\n")
	}
	
	for(var i = 0;i<tracks;i++)
	{
		var t = new Track();
		
		t.setIndex(i);
		t.createClips(clips);
		
		this.tracks[i] = t;
		sendTrackChangeMessage(i,t);
	}
	
	for(var j = 0;j<clips;j++)
	{
		var s = new Scene();
		s.setIndex(j);
		this.scenes[j] = s;
		sendSceneChangeMessage(j,s);
	}
}

Song.prototype.clear = function()
{
	this.tracks = [];
}

Song.prototype.addScene = function(scene)
{
	this.scenes.push(scene);
}

Song.prototype.getScene = function(index)
{
	if(index >= 0 && index < this.scenes.length)
	{
		return this.scenes[index]
	}	
}

Song.prototype.addTrack = function(track)
{
	this.tracks.push(track)
}

Song.prototype.getTrack = function(index)
{
	if(index >= 0 && index < this.tracks.length)
	{
		return this.tracks[index]
	}
}

Song.prototype.getClip = function(trackIndex,clipIndex)
{
	var t = this.getTrack(trackIndex);
	
	if(t != undefined)
	{
		return t.getClip(clipIndex);
	}
	else
	{
		if(DEBUG)
		{
			post("! No track at " + trackIndex + "\n");
		}
	}
}

function Scene()
{
	this.name = "";
	this.tempo = 0;
	this.index = -1;
}

Scene.prototype.getName = function()
{
	return this.name;
}

Scene.prototype.setIndex = function(index)
{
	this.index = index;
}

function Track()
{
	this.clips = [];
	this.name = "";
	this.arm = 0;
    this.can_be_armed = 0;
   	this.mute = 0;
    this.solo = 0;
}

Track.prototype.getClip = function(index)
{
	if(index >= 0 && index < this.clips.length)
	{
		return this.clips[index];
	}
}

Track.prototype.addClip = function(clip)
{
	this.clips.push(clip);
}

Track.prototype.createClips = function(count)
{
	this.clips = [];
	for(var i = 0;i<count;i++)
	{
		var c = new ClipSlot();
		c.setPosition([this.index,i]);
		this.clips[i] = c;
		sendClipChangeMessage(this.index,i,c);
	}
	if(DEBUG)
	{
		post("! Created clips " + this.clips.length + "\n");
	}
}

Track.prototype.getName = function()
{
	return this.name;
}

Track.prototype.setIndex = function(index)
{
	this.index = index;
}

Track.prototype.setTriggeredClip = function(index)
{
	if(DEBUG)
	{
		post("Set Triggered Clip : " + index);
	}
	this.triggeredClip = this.clips[index];
	this.triggeredClip.triggered = true;
}

Track.prototype.clearTriggeredClip = function()
{
	if(this.triggeredClip != undefined)
	{
		if(DEBUG)
		{
			post("Clear Triggered Clip : " + this.triggeredClip + "\n");
		}
		
		this.triggeredClip.triggered = 0;
		var pos = this.triggeredClip.getPosition();
		sendClipChangeMessage(pos[0],pos[1],this.triggeredClip);
		this.triggeredClip = null;
	}
}

// The ClipSlot object
function ClipSlot()
{
	this.name = "";
	this.warped = false;
	this.triggered = false;
	this.playing = false;
	this.loop = false;
	this.audio_or_midi = -1;
}

ClipSlot.prototype.getName = function()
{
	return this.name;
}

ClipSlot.prototype.getPosition = function()
{
	return this.position;
}

ClipSlot.prototype.isEmpty = function()
{
	return this.audio_or_midi == -1;
}

ClipSlot.prototype.isAudio = function()
{
	return this.audio_or_midi == 1;
}

ClipSlot.prototype.isMidi = function()
{
	return this.audio_or_midi == 0;
}

ClipSlot.prototype.getAudioOrMidi = function()
{
	return this.audio_or_midi;
}

ClipSlot.prototype.getPlayState = function()
{
	return [this.playing,this.triggered,this.looping];
}

ClipSlot.prototype.setPosition = function(pos)
{
	this.position = pos;
	this.x = pos[0];
	this.y = pos[1];
}

ClipSlot.prototype.clear = function()
{
	this.name = "";
	this.warped = false;
	this.triggered = false;
	this.playing = false;
	this.loop = false;
	this.audio_or_midi = -1;
}

ClipSlot.prototype.debug = function()
{
	post("ClipSlot at " + this.x + ", " + this.y + " : " + this.col + " : " + " :: " + this.name + "\n")
	post("Playing " + this.playing + "\n")
	post("Triggered " + this.triggered + "\n");
	post("Type : " + (this.audio_or_midi ? "Audio" : "Midi") + "\n");
	post("Mute " + this.mute + "\n");
	post("Warped " + this.warped + "\n")
	post("Looped " + this.loop + "\n")
	post("\n");
}

// GLOBAL CODE
this.song = new Song();
this.song.setSize(4,4);

function HELLO()
{
	post("<< Live says hello\n");
}

function BYE()
{
	post("<< Live says goodbye\n");
}

function BUILD()
{
	post("<< Live says build\n");
	monitorAllClips();
	monitorAllTracks();
	monitorAllScenes();
}

function setSize(tracks,clips)
{
	this.song.setSize(tracks,clips);
}

function getScene(index)
{
	if(DEBUG)
	{
		post(">> Get scene data : " + index + "\n");
	}
	outlet(0,index,GET_SCENE_PARAM,GET_SCENE_CHANNEL);
}

function getAllScenes()
{
	if(DEBUG)
	{
		post(">> Get All Scenes\n");
	}
	var scenes = this.song.scenes;
	for(var s = 0;s<scenes.length;s++)
	{
		getScene(s);
	}
}

function monitorScene(index)
{
	if(DEBUG)
	{
		post(">> Monitor scene : " + index + "\n");
	}
	outlet(0,index,MONITOR_SCENE_PARAM,MONITOR_SCENE_CHANNEL);
}

function stopMonitorScene(index)
{
	if(DEBUG)
	{
		post(">> Stop monitor scene : " + index + "\n");
	}
	outlet(0,index,STOP_MONITOR_SCENE_PARAM,MONITOR_SCENE_CHANNEL);
}

function monitorAllScenes()
{
	var scenes = this.song.scenes;
	for(var s = 0;s<scenes.length;s++)
	{
		monitorScene(s);
	}
}

function stopMonitorAllScenes()
{
	var scenes = this.song.scenes;
	for(var s = 0;s<scenes.length;s++)
	{
		stopMonitorScene(s);
	}
}

function getTrack(index)
{
	if(DEBUG)
	{
		post(">> Get track data : " + index + "\n");
	}
	outlet(0,index,GET_TRACK_PARAM,GET_TRACK_CHANNEL);
}

function getAllTracks()
{
	if(DEBUG)
	{
		post(">> Get All Tracks\n");
	}
	var tracks = this.song.tracks;
	for(var t = 0;t<tracks.length;t++)
	{
		getTrack(t);
	}
}

function monitorTrack(index)
{
	if(DEBUG)
	{
		post(">> Monitor track : " + index + "\n");
	}
	outlet(0,index,MONITOR_TRACK_PARAM,MONITOR_TRACK_CHANNEL);
}

function stopMonitorTrack(index)
{
	if(DEBUG)
	{
		post(">> Stop monitor track : " + index + "\n");
	}
	outlet(0,index,STOP_MONITOR_TRACK_PARAM,MONITOR_TRACK_CHANNEL);
}

function monitorAllTracks()
{
	var tracks = this.song.tracks;
	for(var t = 0;t<tracks.length;t++)
	{
		monitorTrack(t);
	}
}

function stopMonitorAllTracks()
{
	var tracks = this.song.tracks;
	for(var t = 0;t<tracks.length;t++)
	{
		stopMonitorTrack(t);
	}
}

function getClip(track,scene)
{
	if(DEBUG)
	{
		post(">> Get Clip : " + track + ", " + scene + "\n");
	}
	outlet(0,scene,track,GET_CLIP_CHANNEL);
}

function monitorClip(track,scene)
{
	if(DEBUG)
	{
		post(">> Monitor Clip : " + track + ", " + scene + "\n");
	}
	outlet(0,scene,track,MONITOR_CLIP_CHANNEL);
}

function stopMonitorClip(track,scene)
{
	if(DEBUG)
	{
		post(">> Stop Monitoring Clip : " + track + ", " + scene + "\n");
	}
	outlet(0,scene,track,STOP_MONITOR_CLIP_CHANNEL);
}

function getAllClips()
{
	if(DEBUG)
	{
		post(">> Get All Clips\n");
	}
	var tracks = this.song.tracks;
	
	for(var t = 0;t<tracks.length;t++)
	{
		var track = tracks[t];
		for(var c = 0;c<track.clips.length;c++)
		{
			var clip = track.clips[c];
			var pos = clip.getPosition();
			getClip(pos[0],pos[1]);
		}
	}
}

function monitorAllClips()
{
	if(DEBUG)
	{
		post(">> Monitor All Clips\n");
	}
	var tracks = this.song.tracks;
	
	for(var t = 0;t<tracks.length;t++)
	{
		var track = tracks[t];
		for(var c = 0;c<track.clips.length;c++)
		{
			var clip = track.clips[c];
			var pos = clip.getPosition();
			monitorClip(pos[0],pos[1]);
		}
	}
}

function stopMonitorAllClips()
{
	if(DEBUG)
	{
		post(">> Stop monitor All Clips\n");
	}
	var tracks = this.song.tracks;
	
	for(var t = 0;t<tracks.length;t++)
	{
		var track = tracks[t];
		for(var c = 0;c<track.clips.length;c++)
		{
			var clip = track.clips[c];
			var pos = clip.getPosition();
			stopMonitorClip(pos[0],pos[1]);
		}
	}
}

function songData(command, data)
{
	if(DEBUG)
	{
		post("<< Song data " + command + " : " + data + "\n");
	}
	outlet(1,"songChanged",this.song);
}

function sendSceneChangeMessage(sceneIndex,scene)
{
	outlet(4,sceneIndex,"scene",scene);
}

function sendClipChangeMessage(trackIndex,sceneIndex,clipSlot)
{
	outlet(3,trackIndex,sceneIndex,"clip",clipSlot);
}

function sendTrackChangeMessage(trackIndex,track)
{
	outlet(2,trackIndex,"track",track);
}

function sendSongChangeMessage(song)
{
	outlet(1,"song",song);
}

function clipData(cmd)
{
	var command = cmd.command;
	var data = cmd.data;
	
	//post("ClipData " + command + " : " + data + "\n");
	var trackIndex = data.shift();
	var sceneIndex = data.shift();
	
	if(DEBUG)
	{
		post("Clip at " + trackIndex + ", " + sceneIndex + "\n");
	}
	if(this.song == undefined)
	{
		if(DEBUG)
		{
			post("NO SONG!\n");
		}
		return;
	}
	
	clipSlot = this.song.getClip(trackIndex,sceneIndex);
		
	if(clipSlot != undefined)
	{
		if(command == CLIP_DATA)
		{
			clipSlot.col = data.shift()
			// Flags are packed into 6 bits of one byte
			flags = data.shift()
			// decode the flags
			clipSlot.mute = (flags & 32) > 0;
			clipSlot.warped = (flags & 16) > 0;
			clipSlot.loop = (flags & 8) > 0;
			clipSlot.playing = (flags & 4) > 0;
			clipSlot.triggered = (flags & 2) > 0;
			clipSlot.audio_or_midi = (flags & 1);
		
			clipSlot.name = byteToString(data);        			
		}
		else if(command == CLIP_IS_TRIGGERED)
		{
			clipSlot.triggered = data.shift();
			track = this.song.getTrack(trackIndex);
			track.setTriggeredClip(sceneIndex);
		}
		else if(command == CLIP_IS_PLAYING)
		{
			clipSlot.playing = data.shift();
			track = this.song.getTrack(trackIndex);
			track.clearTriggeredClip();
		}
		else if(command == CLIP_NAME)
		{
			clipSlot.name = byteToString(data);
		}
		else if(command == CLIP_IS_LOOPING)
		{
			clipSlot.loop = data.shift();
		}
		else if(command == EMPTY_CLIP)
		{
			clipSlot.clear();
		}
		
		//clipSlot.debug();
		
		sendClipChangeMessage(trackIndex,sceneIndex,clipSlot);
	}
	else
	{
		if(DEBUG)
		{
			post("No ClipSlot at " + trackIndex + ", " + sceneIndex + "\n");
		}
	}
}

function trackData(cmd)
{
	var command = cmd.command;
	var data = cmd.data;
	
	var trackIndex = data.shift();
	var track = this.song.getTrack(trackIndex)
	
	if(track != undefined)
	{
		if(DEBUG)
		{
			post("<< Track Data " + data + "\n");
		}
		if(command == TRACK_DATA)
		{
			flags = data.shift();
			
			track.arm = (flags & 8) > 0;
    		track.can_be_armed = (flags & 4) > 0;
    		track.mute = (flags & 2) > 0;
    		track.solo = (flags & 1);
    		track.name = byteToString(data); 
		}
		else if(command == TRACK_NAME)
		{
			track.name = byteToString(data);
		}
		else if(command == TRACK_IS_ARMED)
		{
			track.arm = data.shift();
		}
		else if(command == TRACK_IS_MUTED)
		{
			track.mute = data.shift();
		}
		else if(command == TRACK_IS_SOLOED)
		{
			track.solo = data.shift();
		}
		
		sendTrackChangeMessage(trackIndex,track);
	}
	else
	{
		if(DEBUG)
		{
			post("No track at " + index + "\n");
		}
	}
}

function sceneData(cmd)
{
	var command = cmd.command;
	var data = cmd.data;
	
	var index = data.shift();
	var scene = this.song.getScene(index)
	
	if(scene != undefined)
	{
		if(DEBUG)
		{
			post("<< Scene Data " + data + "\n");
		}
		if(command == SCENE_DATA)
		{
			scene.name = byteToString(data);
		}
		else if(command == SCENE_NAME)
		{
			scene.name = byteToString(data);
		}
		else if(command == NO_SCENE_DATA)
		{
			scene.name = "";
		}
		
		sendSceneChangeMessage(index,scene);
	}
	else
	{
		if(DEBUG)
		{
			post("No scene at " + index + "\n");
		}	
	}
}

// Convert an array of bytes to a string
function byteToString(byteArray)
{
	var str = "";
	for(i = 0;i<byteArray.length;i++)
	{
		str += String.fromCharCode(byteArray[i]);
	}
	return str;
}
